# Copyright 2023 Cristian Grosu
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from numba import jit, njit, prange

@njit(parallel=True)
def get_nd_indices_parallel(X_nd, metric):
    """ Generates the indices of the nearest neighbors for each point in the nD space.
        Args:
            X_nd (np.ndarray): the nD dataset
            metric (function): the metric to use for calculating the distances
    """
    n_samples = X_nd.shape[0]
    dist_vector = np.zeros(n_samples, dtype=np.float64)
    indices = np.zeros((n_samples, n_samples - 1), dtype=np.int64)
    
    for i in range(n_samples):
        for j in prange(n_samples):
            dist_vector[j] = metric(X_nd[i], X_nd[j])

        indices[i] = np.argsort(dist_vector)[1:] # exclude the point itself
    
    return indices

@njit()
def generate_point_indices(X, point, metric):
    """ Generates the indices of the nearest neighbors for a point.
        Args:
            X (np.ndarray): the (Nd or 2d) dataset
            point (np.ndarray): the (Nd or 2d) point
            metric (function): the metric to use for calculating the distances
    """
    data_samples = X.shape[0]
    dist_vector = np.zeros(data_samples, dtype=np.float64)
        
    for j in range(data_samples):
        dist_vector[j] = metric(point, X[j])

    return np.argsort(dist_vector)
    
@njit(fastmath=True)
def euclidean(x, y):
    r"""Standard euclidean distance.

    ..math::
        D(x, y) = \sqrt{\sum_i (x_i - y_i)^2}
    """
    result = 0.0
    for i in range(x.shape[0]):
        result += (x[i] - y[i]) ** 2
    return np.sqrt(result)

@jit
def get_inv_proj_error(dx, dy):
    """ TODO: add docstring
    """  
    return np.sqrt(np.linalg.norm(dx)**2 + np.linalg.norm(dy)**2)

@njit(parallel=True)
def get_proj_error_parallel(indices_source: np.ndarray, indices_embedding: np.ndarray, k: int=10):
    """ Calculates the projection error for a given data point.
        Args:
            indices_source (np.ndarray): the indices of the point neighbors in the source (i.e. nD) space
            indices_embedding (np.ndarray): the indices of the point neighbors in the embedding (i.e. 2D) space
            k (int): the number of neighbors to consider
    """    
    assert len(indices_source) == len(indices_embedding)
    n = len(indices_source)
    
    continuity = 0.0
    trustworthiness = 0.0
    
    # computing the continuity and trustworthiness errors for each point in parallel
    for i in prange(k):
        rank_2d = 0
        while indices_source[i] != indices_embedding[rank_2d]:
            rank_2d += 1

        rank_nd = 0
        while indices_source[rank_nd] != indices_embedding[i]:
            rank_nd += 1

        if rank_2d > k:
            continuity += rank_2d - k

        if rank_nd > k:
            trustworthiness += rank_nd - k


    continuity = 2 * continuity / (k * (2*n - 3*k - 1))
    trustworthiness = 2 * trustworthiness / (k * (2*n - 3*k - 1))
    
    return (continuity + trustworthiness) / 2

@njit()
def get_proj_error(indices_source: np.ndarray, indices_embedding: np.ndarray, k: int=10):
    """ Calculates the projection error for a given data point.
        Args:
            indices_source (np.ndarray): the indices of the point neighbors in the source (i.e. nD) space
            indices_embedding (np.ndarray): the indices of the point neighbors in the embedding (i.e. 2D) space
            k (int): the number of neighbors to consider
    """    
    assert len(indices_source) == len(indices_embedding)
    n = len(indices_source)
    
    continuity = 0.0
    trustworthiness = 0.0
    
    # computing the continuity and trustworthiness errors for each point in parallel
    for i in range(k):
        rank_2d = 0
        while indices_source[i] != indices_embedding[rank_2d]:
            rank_2d += 1

        rank_nd = 0
        while indices_source[rank_nd] != indices_embedding[i]:
            rank_nd += 1

        if rank_2d > k:
            continuity += rank_2d - k

        if rank_nd > k:
            trustworthiness += rank_nd - k


    continuity = 2 * continuity / (k * (2*n - 3*k - 1))
    trustworthiness = 2 * trustworthiness / (k * (2*n - 3*k - 1))
    
    return (continuity + trustworthiness) / 2


@jit
def get_decode_pixel_priority(img, i, j, window_size, label):
    """Calculates the priority of decoding a chunk of pixels.
       The chunk is defined by the window size and the pixel coordinates (i,j).

    Args:
        img (np.ndarray): the image w.r.t. which the priority is calculated
        i (float): the row of the current pixel
        j (float): the column of the current pixel
        window_size (float): the window size (i.e. the chunk size)
        label (any): the label of the current pixel

    Returns:
        priority (float): the priority of decoding the chunk in range [0,1]
    """
    resolution = img.shape[0]
    # getting the 4 neighbors
    i, j = int(i), int(j)
    w = int(window_size / 2)
    neighbors = []
    if i - w > 0:
        neighbors.append(img[i - w, j])
    if i + w + 1 < resolution:
        neighbors.append(img[i + w + 1, j])
    if j - w > 0 and i + 1 < resolution:
        neighbors.append(img[i + 1, j - w])
    if j + w + 1 < resolution:
        neighbors.append(img[i, j + w + 1])
        
    cost = 0
    for neighbor in neighbors:
        if neighbor != label:
            cost += 1
        
    if cost == 0:
        return -1
    
    cost /= len(neighbors)
    cost *= window_size      
    
    return 1/cost

@njit(parallel=True)
def get_projection_errors_using_inverse_projection( Xnd: np.ndarray, X2d: np.ndarray, spaceNd: np.ndarray, space2d: np.ndarray, progress, k: int=10):  
    
    n_points = len(space2d)                 
    data_samples = len(X2d)
    errors = np.zeros(n_points, dtype=np.float64) 
    
    for index in prange(n_points):              
        indices_embedded, indices_source = np.zeros(data_samples, dtype=np.int64), np.zeros(data_samples, dtype=np.int64)          
        indices_embedded = generate_point_indices(X2d, space2d[index], metric=euclidean)
        indices_source = generate_point_indices(Xnd, spaceNd[index], metric=euclidean)             
        errors[index] = get_proj_error(indices_source, indices_embedded, k=k)          
        progress.update(1)
    
    return errors        



# Copyright 2023 Cristian Grosu
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import matplotlib.pyplot as plt

from ..DBM import SDBM
from ..GUI import DBMPlotterGUI

from .utils import *

def SDBM_usage_example():
    # import the dataset
    X_train, X_test, Y_train, Y_test = import_data()
    
    # import the classifier
    classifier = import_classifier()
    
    # create the DBM
    sdbm = SDBM(classifier=classifier)
    
    # use the SDBM to get the decision boundary map
    img, img_confidence, _, _, _ = sdbm.generate_boundary_map(X_train, Y_train, 
                                                                X_test, Y_test,
                                                                load_folder=os.path.join("tmp", "MNIST", "SDBM"),    
                                                                resolution=256)
    
    
    # make the decision boundary map pretty, by adding the colors and the confidence
    COLORS_MAPPER = {
        -2: [0,0,0], # setting original test data to black
        -1: [1,1,1], # setting original train data to white
        0: [1,0,0], 
        1: [0,1,0], 
        2: [0,0,1], 
        3: [1,1,0], 
        4: [0,1,1], 
        5: [1,0,1], 
        6: [0.5,0.5,0.5], 
        7: [0.5,0,0], 
        8: [0,0.5,0], 
        9: [0,0,0.5]
    }

    color_img = np.zeros((img.shape[0], img.shape[1], 4))
    for i, j in np.ndindex(img.shape):       
        color_img[i,j] = COLORS_MAPPER[img[i,j]] + [img_confidence[i,j]]
    
    # plot the decision boundary map
    plt.title("Decision boundary map")
    plt.axis("off")
    plt.imshow(color_img)
    plt.show()
    
    # use the SDBM to get the inverse projection errors
    img_inverse_projection_errors = sdbm.generate_inverse_projection_errors(resolution=256)
    # plot the inverse projection errors
    fig, ax = plt.subplots()
    ax.set_title("Inverse projection errors")
    ax.xaxis.set_visible(False)
    ax.yaxis.set_visible(False)
    img_ax = ax.imshow(img_inverse_projection_errors)
    fig.colorbar(img_ax, ax=ax)
    plt.show()
    
    # use the SDBM to get the projection errors
    img_projection_errors = sdbm.generate_projection_errors()
    # plot the projection errors
    fig, ax = plt.subplots()
    ax.set_title("Projection errors")
    ax.xaxis.set_visible(False)
    ax.yaxis.set_visible(False)
    img_ax = ax.imshow(img_projection_errors)
    fig.colorbar(img_ax, ax=ax)
    plt.show()
    
def SDBM_usage_example_GUI():
    # import the dataset
    X_train, X_test, Y_train, Y_test = import_data()
    
    Y = np.copy(Y_train)
    
    # use the opf to generate the labels for the training data if the data is not completely labelled
    Y_train = opf(X_train=X_train, Y_train=Y_train)
    
    # generate a classifier
    classifier = generate_classifier(X_train, Y_train)
    
    # import the classifier
    #classifier = import_classifier()
    
    # create the DBM
    sdbm = SDBM(classifier=classifier)
    
    
    # use the DBM to get the decision boundary map, if you don't have the 2D projection of the data
    # the DBM will get it for you, you just need to specify the projection method you would like to use (t-SNE, PCA or UMAP)
    dbm_info = sdbm.generate_boundary_map(X_train, Y, 
                                         X_test, Y_test,
                                         resolution=256,
                                         load_folder=os.path.join("tmp", "MNIST", "SDBM"),                                                                                                                             
                                         )
    
    img, img_confidence, encoded_training_data, encoded_testing_data, training_history  = dbm_info
    
    dbm_plotter_gui = DBMPlotterGUI(dbm_model = sdbm,
                                    img = img,
                                    img_confidence = img_confidence,
                                    encoded_train = encoded_training_data, 
                                    encoded_test = encoded_testing_data,
                                    X_train = X_train,
                                    Y_train = Y_train,
                                    X_test = X_test,
                                    Y_test = Y_test,
                                    save_folder=os.path.join("tmp", "MNIST", "SDBM"), # this is the folder where the DBM will save the changes in data the user makes                                    
                                    )
    dbm_plotter_gui.start()


# Copyright 2023 Cristian Grosu
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import numpy as np
from math import ceil, floor
from queue import PriorityQueue
from scipy import interpolate
import dask.array as da
from sklearn.neighbors import KDTree
from numba_progress import ProgressBar
from tqdm import tqdm

from .tools import get_decode_pixel_priority, get_inv_proj_error, get_nd_indices_parallel, euclidean, get_proj_error_parallel, get_projection_errors_using_inverse_projection
from ..utils import track_time_wrapper
from ..Logger import Logger, LoggerInterface

DBM_DEFAULT_RESOLUTION = 256
time_tracker_console = Logger(name="Decision Boundary Mapper - DBM", info_color="cyan", show_init=False)

class DBMInterface:
    """ Decision Boundary Mapper Interface
    
    Methods to be implemented by the class that implements the DBMInterface:
        fit (X_train, Y_train, X_test, Y_test, epochs=10, batch_size=128, load_folder = None)
        generate_boundary_map (X_train, Y_train, X_test, Y_test, train_epochs=10, train_batch_size=128, resolution=DBM_DEFAULT_RESOLUTION)
        _predict2dspace_ (X)
    
    Example of usage: 
        class SuperBoundaryMapper(DBMInterface):
            def fit (....):
                # To be implemented
            def generate_boundary_map (....):
                # To be implemented
            def _predict2dspace_ (....):
                # To be implemented
    """    
    
    def __init__(self, classifier, logger:LoggerInterface=None):
        """Initializes the classifier and the logger

        Args:
            classifier (Any): The classifier to be used
            logger (LoggerInterface, optional): The logger for the outputting info messages. Defaults to None.
        """
        if logger is None:
            self.console = Logger(name="Decision Boundary Mapper - DBM")
        else:
            self.console = logger
        
        self.console.log("Loaded classifier: " + classifier.name + "")
        self.classifier = classifier
        
    def fit(self, X_train:np.ndarray, Y_train:np.ndarray, 
            X_test:np.ndarray, Y_test:np.ndarray,
            epochs:int=10, batch_size:int=128, load_folder:bool = None):
        """ Trains the classifier on the given data set.

        Args:
            X_train (np.ndarray): Training data set
            Y_train (np.ndarray): Training data labels
            X_test (np.ndarray): Testing data set
            Y_test (np.ndarray): Testing data labels
            epochs (int, optional): The number of epochs for which the DBM is trained. Defaults to 10.
            batch_size (int, optional): Train batch size. Defaults to 128.
        """
        pass
    
    def refit_classifier(self, Xnd:np.ndarray, Y:np.ndarray, save_folder:str, epochs:int=2, batch_size:int=32):
        """ Refits the classifier on the given data set.

        Args:             
            Xnd (np.ndarray): 
            Y (np.ndarray): 
        """
        self.console.log(f"Refiting classifier for {epochs} epochs and batch size {batch_size}, please wait...")
        self.classifier.fit(Xnd, Y, epochs=epochs, batch_size=batch_size, verbose=0)
        self.console.log("Finished refitting classifier")
        self.console.log("Saving a copy of the retrained classifier...")
        self.classifier.save(save_folder, save_format="tf")
        self.console.log("A copy of the retrained classifier was saved!")
    
    def generate_boundary_map(self, 
                              X_train:np.ndarray, Y_train:np.ndarray, 
                              X_test:np.ndarray, Y_test:np.ndarray,
                              train_epochs:int=10, train_batch_size:int=128,
                              resolution:int=DBM_DEFAULT_RESOLUTION,
                              use_fast_decoding:bool=False,
                              projection:str=None
                              ):
        """ Generates a 2D boundary map of the classifier's decision boundary.

        Args:
            X_train (np.ndarray): Training data set
            Y_train (np.ndarray): Training data labels
            X_test (np.ndarray): Testing data set
            Y_test (np.ndarray): Testing data labels
            train_epochs (int, optional): The number of epochs for which the DBM is trained. Defaults to 10.
            train_batch_size (int, optional): Train batch size. Defaults to 128.
            show_predictions (bool, optional): If set to true 10 prediction examples are shown. Defaults to True.
            resolution (int, optional): _description_. Defaults to DBM_DEFAULT_RESOLUTION = 256.
            use_fast_decoding (bool, optional): If set to true the fast decoding algorithm is used. Defaults to False.
            projection (str, optional): The projection to be used for the 2D space. Defaults to None.
        Returns:
            np.array: A 2D numpy array with the decision boundary map
        
        """
        pass
    
    def _predict2dspace_(self, X2d: np.ndarray):
        """ Predicts the labels for the given 2D data set.

        Args:
            X2d (np.ndarray): The 2D data set
        
        Returns:
            predicted_labels (np.array): The predicted labels for the given 2D data set
            predicted_confidences (np.array): The predicted probabilities for the given 2D data set
            spaceNd (np.array): The decoded nD space
        """
        pass
    
    @track_time_wrapper(logger=time_tracker_console)
    def generate_inverse_projection_errors(self, resolution:int, save_folder:str = None):
        """ Calculates the inverse projection errors of the given data.

        Args:
            Xnd (np.array): The nd inverse projection of the data.
        
        Returns:
            errors (np.ndarray): The inverse projection errors matrix of the given data. (resolution x resolution)
        """
            
        self.console.log("Calculating the inverse projection errors of the given data")
        errors = np.zeros((resolution, resolution))
        
        w, h = 1, 1
        
        i = 0
                
        current_row = np.array([(i / resolution, j / resolution) for j in range(resolution) ])
        next_row = np.array([((i + 1) / resolution, j / resolution) for j in range(resolution) ])            
        data_2d = np.concatenate((current_row, next_row))
        data_nd = self.neural_network.decode(data_2d)
        
        previous_row_nd = None
        current_row_nd = data_nd[:resolution]
        next_row_nd = data_nd[resolution:]
        
        
        for i in tqdm(range(resolution)):
            for j in range(resolution):
                dw = w if (j - w < 0) or (j + w >= resolution) else 2 * w
                dh = h if (i - h < 0) or (i + h >= resolution) else 2 * h
                xnd = current_row_nd[j] 
                xl = current_row_nd[j-w] if j - w >= 0 else xnd
                xr = current_row_nd[j+w] if j + w < resolution else xnd
                yl = previous_row_nd[j] if previous_row_nd is not None else xnd
                yr = next_row_nd[j] if next_row_nd is not None else xnd
                                                
                dx = (xl - xr) / dw
                dy = (yl - yr) / dh  
                errors[i,j] = get_inv_proj_error(dx, dy)
            
            previous_row_nd = current_row_nd
            current_row_nd = next_row_nd
            if i + 2 < resolution:
                next_row = np.array([((i + 2) / resolution, j / resolution) for j in range(resolution) ])
                next_row_nd = self.neural_network.decode(next_row)
            else:
                next_row_nd = None
                
        # normalizing the errors to be in the range [0,1]
        errors = (errors - np.min(errors)) / (np.max(errors) - np.min(errors))
        
        if save_folder is not None:
            self.console.log("Saving the inverse projection errors results")
            save_path = os.path.join(save_folder, "inverse_projection_errors.npy")
            with open(save_path, "wb") as f:
                np.save(f, errors)
            self.console.log("Saved inverse projection errors results!")
            
        return errors
    
    @track_time_wrapper(logger=time_tracker_console)
    def _get_img_dbm_(self, resolution:int):
        """ This function generates the 2D image of the boundary map using the trained autoencoder and classifier.

        Args:
            resolution (int): The resolution of the 2D image to be generated 

        Returns:
            img (np.array): The 2D image of the boundary map
            img_confidence (np.array): The confidence of map of each pixel of the 2D image of the boundary map
            img_space_Nd (np.array): The nD space points
      
        Example:
            >>> img, img_confidence = self._get_img_dbm_(resolution = 100)
        """
        space2d = np.array([(i / resolution, j / resolution) for i in range(resolution) for j in range(resolution)])
        self.console.log("Predicting labels for the 2D boundary mapping using the nD data and the trained classifier...")
        
        chunk_size = 10000
        chunks = (resolution * resolution) // chunk_size + 1        
        space2d_chunks = np.array_split(space2d, chunks)
        
        img, img_confidence = np.array([]), np.array([])
        
        chunk_index = 0
        for space2d_chunk in space2d_chunks:
            chunk_index += 1
            self.console.log(f"Predicting labels for the 2D boundary mapping using the nD data and the trained classifier... (chunk {chunk_index}/{len(space2d_chunks)})")
            predicted_labels, predicted_confidence = self._predict2dspace_(space2d_chunk)
            img = np.concatenate((img, predicted_labels))
            img_confidence = np.concatenate((img_confidence, predicted_confidence))
        
        img = img.reshape((resolution, resolution))
        img_confidence = img_confidence.reshape((resolution, resolution))        
        
        return (img, img_confidence)
    
    @track_time_wrapper(logger=time_tracker_console)
    def _get_img_dbm_fast_(self, resolution:int, computational_budget=None, interpolation_method:str="linear"):
        """
        This function generates the 2D image of the boundary map. It uses a fast algorithm to generate the image.
        
        ATTENTION: Highly recommended to use resolution = 2 ^ n, where n is an integer number (powers of 2).
        
        Args:
            resolution (int): the resolution of the 2D image to be generated
            computational_budget (int, optional): The computational budget to be used. Defaults to None.
            interpolation_method (str, optional): The interpolation method to be used for the interpolation of sparse data generated by the fast algorithm.
                                                  Defaults to "linear". The options are: "nearest", "linear", "cubic"
        Returns:
            img, img_confidence: The 2D image of the boundary map and a image with the confidence for each pixel
            spaceNd: The nD space points
        Example:
            >>> img, img_confidence = self._get_img_dbm_fast_(resolution=32, computational_budget=1000)
        """
        # ------------------------------------------------------------
        # Setting the initial parameters
        WINDOW_SIZE = 8
        NEIGHBORS_NUMBER = 4
        
        assert(resolution > WINDOW_SIZE)
        
        INITIAL_RESOLUTION = resolution // WINDOW_SIZE
        
        if computational_budget is None:
            computational_budget = resolution * resolution
        
        assert(computational_budget > INITIAL_RESOLUTION * INITIAL_RESOLUTION)
        INITIAL_COMPUTATIONAL_BUDGET = computational_budget
        
        if (resolution % INITIAL_RESOLUTION != 0):
            self.console.warn(f"The required resolution is not a multiple of the initial window size ({WINDOW_SIZE} x {WINDOW_SIZE})")
            self.console.log("The resolution will be set to the closest multiple of the window size")
            resolution = INITIAL_RESOLUTION * WINDOW_SIZE
            self.resolution = resolution   
            self.console.log(f"Resolution was set to {resolution} x {resolution}")
            
        window_size = WINDOW_SIZE
        # ------------------------------------------------------------
        
        # ------------------------------------------------------------
        img = np.zeros((resolution, resolution))
        priority_queue = PriorityQueue()
        # ------------------------------------------------------------
                
        # ------------------------------------------------------------
        # generate the initial points
        self.console.log(f"Generating the initial central points within each window... total number of windows ({(INITIAL_RESOLUTION * INITIAL_RESOLUTION)})")
        
        space2d = [((i * window_size + window_size / 2 - 0.5) / resolution, (j * window_size + window_size / 2 - 0.5) / resolution) for i in range(INITIAL_RESOLUTION) for j in range(INITIAL_RESOLUTION)]
        
        predicted_labels, predicted_confidence = self._predict2dspace_(space2d)
        #shape = (INITIAL_RESOLUTION, INITIAL_RESOLUTION, ) + predicted_spaceNd.shape[1:]        
        #spaceNd = predicted_spaceNd.reshape(shape)
        
        predictions = predicted_labels.reshape((INITIAL_RESOLUTION, INITIAL_RESOLUTION))
        confidences = predicted_confidence.reshape((INITIAL_RESOLUTION, INITIAL_RESOLUTION))        
        
        # initialize the nD space
        
        computational_budget -= INITIAL_RESOLUTION * INITIAL_RESOLUTION
        
        
        #shape = (resolution, resolution, ) + spaceNd.shape[2:]
        #img_space_Nd = np.zeros(shape, dtype=np.float32)
                
        # fill the initial points in the 2D image
        for i in range(INITIAL_RESOLUTION):
            for j in range(INITIAL_RESOLUTION):
                x0, x1, y0, y1 = i * window_size, (i+1) * window_size, j * window_size, (j+1) * window_size
                img[x0:x1, y0:y1] = predictions[i,j]
                #img_space_Nd[x0:x1, y0:y1] = spaceNd[i,j]
        # -------------------------------------
        
        # collecting the indexes of the actual computed pixels in the 2D image and the confidence of each pixel
        # creating an artificial border for the 2D confidence image
        confidence_map = []
        
        if interpolation_method != "nearest":
            border_indices = [(i * window_size + window_size / 2 - 0.5, 0) for i in range(INITIAL_RESOLUTION)] + \
                            [(i * window_size + window_size / 2 - 0.5, resolution - 1) for i in range(INITIAL_RESOLUTION)] + \
                            [(0, i * window_size + window_size / 2 - 0.5) for i in range(-1, INITIAL_RESOLUTION+1)] + \
                            [(resolution - 1, i * window_size + window_size / 2 - 0.5) for i in range(-1, INITIAL_RESOLUTION + 1)]
            
            space2d_border = [(i / resolution, j / resolution) for (i,j) in border_indices]
            _, confidences_border = self._predict2dspace_(space2d_border)
            computational_budget -= len(space2d_border)
            confidence_map = [(i, j, conf) for (i,j),conf in zip(border_indices, confidences_border)]
                    
        # analyze the initial points and generate the priority queue        
        for i in range(INITIAL_RESOLUTION):
            for j in range(INITIAL_RESOLUTION):
                x, y = i * window_size + window_size / 2 - 0.5, j * window_size + window_size / 2 - 0.5
                confidence_map.append((x,y,confidences[i,j]))                
                priority = get_decode_pixel_priority(img, x, y, window_size, predictions[i,j])
                if priority != -1:
                    priority_queue.put((priority, (window_size, x, y)))

        # -------------------------------------
        # start the iterative process of filling the image
        self.console.log(f"Starting the iterative process of refining windows...")
        while computational_budget > 0 and not priority_queue.empty():    
            # take the highest priority task
            priority, item = priority_queue.get()
            
            # getting all the items with the same priority
            items = [item]
            if not priority_queue.empty():
                next_priority, next_item = priority_queue.get()
                while priority == next_priority:
                    items.append(next_item)
                    if priority_queue.empty():
                        break
                    next_priority, next_item = priority_queue.get()
                if priority != next_priority:
                    priority_queue.put((next_priority, next_item))
            
            
            space2d, indices = [], []
            single_points_space, single_points_indices = [], []
            valid_items = []
            
            for item in items:            
                (window_size, i, j) = item
                
                if window_size == 1:
                    single_points_indices += [(int(i), int(j))]
                    single_points_space += [(i / resolution, j / resolution)]
                    continue
                
                window_size = window_size / NEIGHBORS_NUMBER
                neighbors = [(i - window_size, j - window_size), (i - window_size, j + window_size), (i + window_size, j - window_size), (i + window_size, j + window_size)]
                space2d += [(x / resolution, y / resolution) for (x, y) in neighbors]
                indices += neighbors                                                
                valid_items.append(item)
            
            
            space = space2d + single_points_space
            
            # check if the computational budget is enough and update it
            if computational_budget - len(space) < 0:
                self.console.warn("Computational budget exceeded, stopping the process")
                break
            
            computational_budget -= len(space)
            
            # decode the space
            predicted_labels, predicted_confidence = self._predict2dspace_(space)
            # copy the image to a new one, the new image will be updated with the new labels, the old one will be used to calculate the priorities
            new_img = np.copy(img)            
            
            # fill the new image with the single points
            single_points_labels = predicted_labels[len(space2d):]
            single_points_confidences  = predicted_confidence[len(space2d):]
            #single_points_spaceNd = predicted_spaceNd[len(space2d):]
            
            for i in range(len(single_points_indices)):
                new_img[single_points_indices[i]] = single_points_labels[i]
                #img_space_Nd[single_points_indices[i]] = single_points_spaceNd[i]
                confidence_map.append((single_points_indices[i][0], single_points_indices[i][1], single_points_confidences[i]))                
                           
            predicted_labels = predicted_labels[:len(space2d)]
            predicted_confidence = predicted_confidence[:len(space2d)]
            #predicted_spaceNd = predicted_spaceNd[:len(space2d)] 
            
            # fill the new image with the new labels and update the priority queue
            for index in range(len(valid_items)):
                (window_size, i, j) = valid_items[index]
                neighbors = indices[index*NEIGHBORS_NUMBER:(index+1)*NEIGHBORS_NUMBER]
                neighbors_labels = predicted_labels[index*NEIGHBORS_NUMBER:(index+1)*NEIGHBORS_NUMBER]
                confidences = predicted_confidence[index*NEIGHBORS_NUMBER:(index+1)*NEIGHBORS_NUMBER]
                #spaceNd = predicted_spaceNd[index*NEIGHBORS_NUMBER:(index+1)*NEIGHBORS_NUMBER]
                
                new_window_size = window_size / NEIGHBORS_NUMBER
                
                for (x, y), label, conf in zip(neighbors, neighbors_labels, confidences):
                    # all the pixels in the window are set to the same label
                    # starting from the top left corner of the window
                    # ending at the bottom right corner of the window
                    # the +1 is because the range function is not inclusive
                    confidence_map.append((x, y, conf))
                    if new_window_size >= 1:
                        x0, x1, y0, y1 = ceil(x-new_window_size), floor(x + new_window_size), ceil(y - new_window_size), floor(y + new_window_size)
                        new_img[x0:x1 + 1, y0:y1 + 1] = label                    
                        #img_space_Nd[x0:x1 + 1, y0:y1 + 1] = pointNd
                    else:
                        new_img[int(x), int(y)] = label
                        #img_space_Nd[int(x), int(y)] = pointNd
                
                # update the priority queue after the window has been filled
                for (x, y), label in zip(neighbors, neighbors_labels):
                    priority = get_decode_pixel_priority(img, x, y, 2 * new_window_size, label)
                    if priority != -1:
                        priority_queue.put((priority, (2 * new_window_size, x, y)))
            
            # update the image        
            img = new_img
        
        # summary
        self.console.log(f"Finished decoding the image, initial computational budget: {INITIAL_COMPUTATIONAL_BUDGET} computational budget left: {computational_budget}")
        self.console.log(f"Items left in the priority queue: {priority_queue.qsize()}")
        
        # generating the confidence image using interpolation based on the confidence map
        img_confidence = self._generate_interpolated_image_(sparse_map=confidence_map,
                                                            resolution=resolution,
                                                            method=interpolation_method).T
        

        return img, img_confidence

    def _generate_interpolated_image_(self, sparse_map, resolution, method='linear'):
        """A private method that uses interpolation to generate the values for the 2D space image
           The sparse map is a list of tuples (x, y, data)
           The sparse map represents a structured but non uniform grid of data values
           Therefore usual rectangular interpolation methods are not suitable
           For the interpolation we use the scipy.interpolate.griddata function with the linear method
        Args:
            sparse_map (list): a list of tuples (x, y, data) where x and y are the coordinates of the pixel and data is the data value
            resolution (int): the resolution of the image we want to generate (the image will be a square image)
            method (str, optional): The method to be used for the interpolation. Defaults to 'linear'. Available methods are: 'nearest', 'linear', 'cubic'

        Returns:
            np.array: an array of shape (resolution, resolution) containing the data values for the 2D space image
        """
        X, Y, Z = [], [], []
        for (x, y, z) in sparse_map:
            X.append(x)
            Y.append(y)
            Z.append(z)
        X, Y, Z = np.array(X), np.array(Y), np.array(Z)
        xi = np.linspace(0, resolution-1, resolution)
        yi = np.linspace(0, resolution-1, resolution)
        return interpolate.griddata((X, Y), Z, (xi[None,:], yi[:,None]), method=method)

    @track_time_wrapper(logger=time_tracker_console)
    def _generate_interpolation_rbf_(self, sparse_map, resolution, function='linear'):
        """A private method that uses interpolation to generate the values for the 2D space image
           The sparse map is a list of tuples (x, y, data) where x, y and data are in the range [0, 1]
        
        Args:
            sparse_map (np.ndarray): a list of tuples (x, y, data) where x and y are the coordinates of the pixel and data is the data value
            resolution (int): the resolution of the image we want to generate (the image will be a square image)
            function (str, optional): Defaults to 'euclidean'.
        """
        self.console.log("Computing the interpolated image using RBF interpolation...")
        X, Y, Z = [], [], []
        for (x, y, z) in sparse_map:
            X.append(x)
            Y.append(y)
            Z.append(z)

        rbf = interpolate.Rbf(X, Y, Z, function=function) 
        ti = np.linspace(0, 1, resolution)
        xx, yy = np.meshgrid(ti, ti)
        # using dask to parallelize the computation of the rbf function
        # the rbf function is applied to each pixel of the image
        # the image is divided into chunks and each chunk is processed in parallel
        # the result is then merged together
        # the number of chunks is equal to the number of cores
        cores = 4
        ix = da.from_array(xx, chunks=(1, cores))
        iy = da.from_array(yy, chunks=(1, cores))
        iz = da.map_blocks(rbf, ix, iy)
        zz = iz.compute()
        self.console.log("Finished computing the interpolated image using RBF interpolation")
        return zz
        
    def generate_projection_errors(self, Xnd: np.ndarray = None, 
                                   X2d: np.ndarray = None, 
                                   resolution: int = None, 
                                   use_interpolation:bool=True,
                                   save_folder:str = None):
        """ Calculates the projection errors of the given data.

        Args:
            Xnd (np.array): The data to be projected. The data must be in the range [0,1].
            X2d (np.array): The 2D projection of the data. The data must be in the range [0,1].
            resolution (int): The resolution of the 2D space.
            spaceNd (np.array): The nD space of the data.
            use_interpolation (bool): Whether to use interpolation to generate the projection errors or use the inverse projection. Defaults to interpolation usage
            save_folder (str): The folder path where to store the projection errors results. Defaults to None.
        Returns:
            errors (np.array): The projection errors matrix of the given data. 
        
        Example:
            >>> from decision-boundary-mapper import SDBM
            >>> classifier = ...
            >>> Xnd = ...
            >>> X2d = ...
            >>> sdbm = SDBM(classifier)
            >>> errors = sdbm.get_projection_errors(Xnd, X2d)
            >>> plt.imshow(errors)
            >>> plt.show()
            >>> ...
            >>> spaceNd = ...
            >>> errors = sdbm.get_projection_errors(Xnd, X2d, spaceNd=spaceNd, use_interpolation=False)
            >>> plt.imshow(errors)
            >>> plt.show()
        """
        
        if Xnd is None:
            if self.Xnd is None:
                self.console.error("No nD data provided and no data stored in the DBM object.")
                raise ValueError("No nD data provided and no data stored in the DBM object.")
            Xnd = self.Xnd
        if X2d is None:
            if self.X2d is None:
                self.console.error("No 2D data provided and no data stored in the DBM object.")
                raise ValueError("No 2D data provided and no data stored in the DBM object.")
            X2d = self.X2d
        if resolution is None:
            if self.resolution is None:   
                self.console.error("The resolution of the 2D space is not set, try to call the method 'generate_boundary_map' first.")
                raise Exception("The resolution of the 2D space is not set, try to call the method 'generate_boundary_map' first.")
            resolution = self.resolution
        
        assert len(X2d) == len(Xnd)    
        X2d = X2d.reshape((X2d.shape[0], -1))
        Xnd = Xnd.reshape((Xnd.shape[0], -1))
        
        assert X2d.shape[1] == 2
        
        self.console.log("Calculating the projection errors of the given data, this might take a couple of minutes. Please wait...")
        if use_interpolation:
            errors = self._generate_projection_errors_using_interpolation_(Xnd, X2d, resolution)                    
        else:            
            errors = self._generate_projection_errors_using_inverse_projection_(Xnd, X2d, resolution)
        
        self.console.log("Finished computing the projection errors!")
        if save_folder is not None:
            self.console.log("Saving the projection errors results")
            save_path = os.path.join(save_folder, "projection_errors_interpolated.npy") if use_interpolation else os.path.join(save_folder, "projection_errors_inv_proj.npy")
            with open(save_path, "wb") as f:
                np.save(f, errors)
            self.console.log("Saved projection errors results!")
            
        return errors
    
    @track_time_wrapper(logger=time_tracker_console)    
    def _generate_projection_errors_using_interpolation_(self, Xnd, X2d, resolution):  
        errors = np.zeros((resolution,resolution))  
        K = 10 # Number of nearest neighbors to consider when computing the errors
        metric = "euclidean"
        
        self.console.log("Computing the 2D tree")
        tree = KDTree(X2d, metric=metric)
        self.console.log("Finished computing the 2D tree")
        self.console.log("Computing the 2D tree indices")
        indices_embedded = tree.query(X2d, k=len(X2d), return_distance=False)
        # Drop the actual point itself
        indices_embedded = indices_embedded[:, 1:]
        self.console.log("Finished computing the 2D tree indices")
        
        self.console.log("Calculating the nD distance indices")
        indices_source = get_nd_indices_parallel(Xnd, metric=euclidean)
        self.console.log("Finished computing the nD distance indices")
        
        sparse_map = []
        for k in range(len(X2d)):
            x, y = X2d[k]
            sparse_map.append( (x, y, get_proj_error_parallel(indices_source[k], indices_embedded[k], k=K)))
            
        errors = self._generate_interpolation_rbf_(sparse_map, resolution, function='linear').T
        
        # resize the errors in range [0,1]
        errors = (errors - errors.min()) / (errors.max() - errors.min())
        
        return errors
    
    @track_time_wrapper(logger=time_tracker_console)
    def _generate_projection_errors_using_inverse_projection_(self, 
                                                            Xnd: np.ndarray = None, 
                                                            X2d: np.ndarray = None,                                                             
                                                            resolution: int = 256):        
        
        K = 10 # Number of nearest neighbors to consider when computing the errors       
        space2d = np.array([(i / resolution, j / resolution) for i in range(resolution) for j in range(resolution)]) # generate the 2D flatten space
        
        errors = np.array([])
        # split space2d into chunks
        chunks_number = resolution * resolution // 10000 + 1 # split the space into chunks of 10000 points max
        self.console.log(f"Splitting the 2D space into {chunks_number} chunks")
        space2d_chunks = np.array_split(space2d, chunks_number)
        
        chunk_index = 0
        for space2d_chunk in space2d_chunks:
            chunk_index += 1
            self.console.log(f"Computing the projection errors for chunk ({chunk_index}/{chunks_number}) of size: {(len(space2d_chunk))}")
            spaceNd_chunk = self.neural_network.decode(space2d_chunk) # decode the 2D space to nD space        
            spaceNd_chunk = spaceNd_chunk.reshape((spaceNd_chunk.shape[0], -1)) # flatten the space   
                    
            with ProgressBar(total=len(space2d_chunk)) as progress:
                errors_chunk = get_projection_errors_using_inverse_projection(Xnd=Xnd, X2d=X2d, spaceNd=spaceNd_chunk, space2d=space2d_chunk, k=K, progress=progress)
        
            errors = np.concatenate((errors, errors_chunk))
            
        # resize the errors in range [0,1]
        errors = (errors - errors.min()) / (errors.max() - errors.min())        
        errors = errors.reshape((resolution, resolution))        
        
        return errors


# Copyright 2023 Cristian Grosu
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2023 Cristian Grosu
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from math import sqrt
from datetime import datetime
from matplotlib.patches import Patch, Circle
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk
from matplotlib.offsetbox import OffsetImage, AnnotationBbox, TextArea
import matplotlib.figure as figure
import matplotlib.pyplot as plt
from PIL import Image
import numpy as np
import PySimpleGUI as sg
import os

from .. import Logger, LoggerGUI

def draw_figure_to_canvas(canvas, figure, canvas_toolbar):
    if canvas.children:
        for child in canvas.winfo_children():
            child.destroy()
    if canvas_toolbar.children:
        for child in canvas_toolbar.winfo_children():
            child.destroy()
    figure_canvas_agg = FigureCanvasTkAgg(figure, canvas)
    figure_canvas_agg.draw()
    toolbar = NavigationToolbar2Tk(figure_canvas_agg, canvas_toolbar)        
    toolbar.update()  
    figure_canvas_agg.get_tk_widget().pack(side='top', fill='both', expand=1)
    return figure_canvas_agg

def generate_color_mapper():
    colors_mapper = {
        -2: [0,0,0], # setting original test data to black
        -1: [1,1,1], # setting original train data to white
        0: [1,0,0], 
        1: [0,1,0], 
        2: [0,0,1], 
        3: [1,1,0], 
        4: [0,1,1], 
        5: [1,0,1], 
        6: [0.5,0.5,0.5], 
        7: [0.5,0,0], 
        8: [0,0.5,0], 
        9: [0,0,0.5]
    }
    # setting the rest of the colors 
    for i in range(10, 100):
        colors_mapper[i] = [i/200,i/200,i/200]
    return colors_mapper

# Generating initial settings
COLORS_MAPPER = generate_color_mapper()
APP_FONT = 'Helvetica 12'
TITLE = "Decision Boundary Map & Errors"
WINDOW_SIZE = (1550, 950)
BLACK_COLOR = "#252526"
BUTTON_PRIMARY_COLOR = "#007acc"
WHITE_COLOR = "#ffffff"
RIGHTS_MESSAGE_1 = "© 2023 Cristian Grosu. All rights reserved."
RIGHTS_MESSAGE_2 = "Made by Cristian Grosu for Utrecht University Master Thesis in 2023"
INFORMATION_CONTROLS_MESSAGE = "To change label(s) of a data point click on the data point, or select the data point by including them into a circle.\nPress any digit key to indicate the new label.\nPress 'Enter' to confirm the new label. Press 'Esc' to cancel the action.\nTo remove a change just click on the data point.\nAfter the changes are done press 'Apply Changes' to update the model. \nAfter the changes are applied the window will update."
DBM_WINDOW_ICON_PATH = os.path.join(os.path.dirname(__file__), "assets", "dbm_plotter_icon.png")
CLASSIFIER_PERFORMANCE_HISTORY_FILE = "classifier_performance.log"
LABELS_CHANGES_FILE = "label_changes.json"

class DBMPlotterGUI:
    def __init__ (self, 
                  dbm_model,
                  img, img_confidence,
                  X_train, Y_train, 
                  X_test, Y_test,
                  encoded_train, encoded_test,                                    
                  save_folder,
                  projection_technique=None,
                  logger=None,
                  main_gui=None,
                  ):        
        """[summary] DBMPlotterGUI is a GUI that allows the user to visualize the decision boundary map and the errors of the DBM model.
        It also allows the user to change the labels of the data points and see the impact of the changes on the model.

        Args:
            dbm_model (DBM or SDBM): The DBM model that will be used to generate the decision boundary map.
            img (np.ndarray): The decision boundary map image.
            img_confidence (np.ndarray): The decision boundary map confidence image.
            X_train (np.ndarray): The training data.
            Y_train (np.ndarray): The training labels.
            X_test (np.ndarray): The test data.
            Y_test (np.ndarray): The test labels.
            encoded_train (np.ndarray): Positions of the training data points in the 2D embedding space.
            encoded_test (np.ndarray): Positions of the test data points in the 2D embedding space.
            spaceNd (np.ndarray): This is a list of lists of size resolution*resolution. Each point in this list is an nd data point which can be a vector, a matrix or a multidimensional matrix.
            save_folder (string): The folder where all the DBM model related files will be saved.
            projection_technique (string, optional): The projection technique the user wants to use if DBM is used as dbm_model. Defaults to None.
            logger (Logger, optional): The logger which is meant for logging the info messages. Defaults to None.
            main_gui (GUI, optional): The GUI that started the DBMPlotterGUI if any. Defaults to None.
        """
        if logger is None:
            self.console = Logger(name="DBMPlotterGUI")
        else:
            self.console = logger
        
        self.main_gui = main_gui # reference to main window  
        
        self.positions_of_labels_changes = ([], [], [])
        self.inverse_projection_errors = None
        self.projection_errors = None     
        
        self.initialize(dbm_model, 
                        img, 
                        img_confidence, 
                        X_train, Y_train, 
                        X_test, Y_test, 
                        encoded_train, encoded_test, 
                        save_folder, 
                        projection_technique)
    
    def initialize(self,
                   dbm_model,
                   img, img_confidence,
                   X_train, Y_train, 
                   X_test, Y_test,
                   encoded_train, encoded_test,                   
                   save_folder,
                   projection_technique=None,                   
                   ):
        self.dbm_model = dbm_model 
        self.img = img
        self.img_confidence = img_confidence
        self.X_train = X_train
        self.Y_train = Y_train
        self.X_test = X_test
        self.Y_test = Y_test
        self.encoded_train = encoded_train
        self.encoded_test = encoded_test  
        self.save_folder = save_folder # folder where to save the changes made to the data by the user    
        self.projection_technique = projection_technique # projection technique used to generate the DBM                                    
        self.color_img, self.legend = self._build_2D_image_(img)
        self.train_mapper, self.test_mapper = self._generate_encoded_mapping_()
        
        # --------------------- Plotter related ---------------------                
        self.fig, self.ax = self._build_plot_()        
        self._build_annotation_mapper_()
        
        self.fig.legend(handles=self.legend, borderaxespad=0. )
        # --------------------- Plotter related ---------------------
        
        # --------------------- Others ------------------------------           
        self.motion_event_cid = None
        self.click_event_cid = None
        self.key_event_cid = None
        self.release_event_cid = None
        self.current_selected_point = None
        self.current_selected_point_assigned_label = None
        self.expert_updates_labels_mapper = {}        
        self.stop_application = False 
        self.update_labels_by_circle_select = True 
        self.update_labels_circle = None  
        
    def _initialize_gui_(self):        
        # --------------------- GUI related ---------------------
        self.window = self._build_GUI_()
        self.canvas, self.fig_agg, self.canvas_controls = self._build_canvas_(self.fig, key = "-DBM CANVAS-", controls_key="-CONTROLS CANVAS-")
        
        # --------------------- Classifier related ---------------------
        self.compute_classifier_metrics()
        
        self.draw_dbm_img()            
        # --------------------- GUI related ---------------------        
        self.updates_logger = LoggerGUI(name = "Updates logger", output = self.window["-LOGGER-"], update_callback = self.window.refresh)
        self.dbm_model.console = self.updates_logger
   
    def _get_GUI_layout_(self):   
        buttons_proj_errs = []
        buttons_inv_proj_errs = []        
        computed_projection_errors = self.projection_errors is not None
        computed_inverse_projection_errors = self.inverse_projection_errors is not None
        if not computed_projection_errors: 
            buttons_proj_errs = [
                sg.Button('Compute Projection Errors (interpolation)', font=APP_FONT, expand_x=True, key="-COMPUTE PROJECTION ERRORS INTERPOLATION-", button_color=(WHITE_COLOR, BUTTON_PRIMARY_COLOR)),
                sg.Button('Compute Projection Errors (inverse projection)', font=APP_FONT, expand_x=True, key="-COMPUTE PROJECTION ERRORS INVERSE PROJECTION-", button_color=(WHITE_COLOR, BUTTON_PRIMARY_COLOR))             
            ]
        if not computed_inverse_projection_errors:
            buttons_inv_proj_errs = [
                sg.Button('Compute Inverse Projection Errors', font=APP_FONT, expand_x=True, key="-COMPUTE INVERSE PROJECTION ERRORS-", button_color=(WHITE_COLOR, BUTTON_PRIMARY_COLOR)) 
            ]        
         
        layout = [                                  
                    [sg.Canvas(key='-DBM CANVAS-', expand_x=True, expand_y=True, pad=(0,0))],     
                    [sg.Canvas(key='-CONTROLS CANVAS-', expand_x=True, pad=(0,0))],
                    [
                        sg.Column([
                            [
                                sg.Button("Show classifier performance history", font=APP_FONT, expand_x=True, key="-SHOW CLASSIFIER PERFORMANCE HISTORY-", button_color=(WHITE_COLOR, BUTTON_PRIMARY_COLOR)),
                                sg.Text("Classifier accuracy: ", font=APP_FONT, expand_x=True, key="-CLASSIFIER ACCURACY-"),  
                            ],
                            [  
                                sg.Checkbox("Change labels by selecting with circle", default=True, key="-CIRCLE SELECTING LABELS-", enable_events=True, font=APP_FONT, expand_x=True, pad=(0,0)),
                                sg.Checkbox("Show labels changes", default=False, key="-SHOW LABELS CHANGES-", enable_events=True, font=APP_FONT, expand_x=True, pad=(0,0)),
                            ],
                            [
                                sg.Checkbox("Show dbm color map", default=True, key="-SHOW DBM COLOR MAP-", enable_events=True, font=APP_FONT, expand_x=True, pad=(0,0)),
                                sg.Checkbox("Show dbm confidence", default=True, key="-SHOW DBM CONFIDENCE-", enable_events=True, font=APP_FONT, expand_x=True, pad=(0,0)),                                
                            ],
                            [
                                sg.Checkbox("Show inverse projection errors", default=False, key="-SHOW INVERSE PROJECTION ERRORS-", enable_events=True, font=APP_FONT, expand_x=True, pad=(0,0), visible=computed_inverse_projection_errors),
                                sg.Checkbox("Show projection errors", default=False, key="-SHOW PROJECTION ERRORS-", enable_events=True, font=APP_FONT, expand_x=True, pad=(0,0), visible=computed_projection_errors),
                            ],
                            #[
                            #    sg.Text("Epochs to train: ", font=APP_FONT, expand_x=True, pad=(0,0), key="-EPOCHS LABEL-"),
                            #    sg.Input("2", font=APP_FONT, key="-EPOCHS-", background_color=WHITE_COLOR, text_color=BLACK_COLOR, expand_x=True, justification="center"),                            
                            #],
                            [
                                sg.Button('Apply Updates', font=APP_FONT, expand_x=True, key="-APPLY CHANGES-", button_color=(WHITE_COLOR, BUTTON_PRIMARY_COLOR)),
                            ],
                            buttons_proj_errs,
                            buttons_inv_proj_errs,
                            [sg.Text(INFORMATION_CONTROLS_MESSAGE, expand_x=True)],
                            [sg.Text(RIGHTS_MESSAGE_1, expand_x=True)],
                            [sg.Text(RIGHTS_MESSAGE_2, expand_x=True)],           
                        ], expand_x=True),
                        sg.Column([
                            [sg.Multiline("", key="-LOGGER-", size=(50,20), background_color=WHITE_COLOR, text_color=BLACK_COLOR, auto_size_text=True, expand_y=True, expand_x=True)],                    
                        ], expand_x=True),
                    ]                                                      
                ]
        
        return layout
        
    def _build_GUI_(self):        
        window = sg.Window(title = TITLE, 
                           layout = self._get_GUI_layout_(), 
                           size=WINDOW_SIZE, 
                           resizable=True,
                           icon=DBM_WINDOW_ICON_PATH,
                           element_justification='center',
                           )
        window.finalize()
        #window.maximize()
        return window
    
    def start(self):
        self._initialize_gui_()
        while True:
            event, values = self.window.read()
            
            if event == "Exit" or event == sg.WIN_CLOSED or self.stop_application:
                break
        
            self.handle_event(event, values)            
        
        self.stop()
    
    def stop(self):
        if self.main_gui is not None and hasattr(self.main_gui, "handle_changes_in_dbm_plotter"):
            self.main_gui.handle_changes_in_dbm_plotter()
        self.console.log("Closing the application...")
        
        classifier_performance_path = os.path.join(self.save_folder, CLASSIFIER_PERFORMANCE_HISTORY_FILE)
        labels_changes_path = os.path.join(self.save_folder, LABELS_CHANGES_FILE)
        
        if os.path.exists(classifier_performance_path):
            os.remove(classifier_performance_path)
        
        if os.path.exists(labels_changes_path):
            os.remove(labels_changes_path)
                    
        self.window.close()
    
    def handle_event(self, event, values):
        EVENTS = {
            "-COMPUTE PROJECTION ERRORS INTERPOLATION-": self.handle_compute_projection_errors_event,
            "-COMPUTE PROJECTION ERRORS INVERSE PROJECTION-": self.handle_compute_projection_errors_event,
            "-COMPUTE INVERSE PROJECTION ERRORS-": self.handle_compute_inverse_projection_errors_event,
            "-APPLY CHANGES-": self.handle_apply_changes_event,
            "-SHOW DBM COLOR MAP-": self.handle_checkbox_change_event,
            "-SHOW DBM CONFIDENCE-": self.handle_checkbox_change_event,
            "-SHOW INVERSE PROJECTION ERRORS-": self.handle_checkbox_change_event,
            "-SHOW PROJECTION ERRORS-": self.handle_checkbox_change_event,
            "-SHOW LABELS CHANGES-": self.handle_checkbox_change_event,
            "-CIRCLE SELECTING LABELS-": self.handle_circle_selecting_labels_change_event,            
            "-SHOW CLASSIFIER PERFORMANCE HISTORY-": self.handle_show_classifier_performance_history_event,
        }
        
        EVENTS[event](event, values)    
    
    def _build_2D_image_(self, img, class_name_mapper = lambda x: str(x), colors_mapper = COLORS_MAPPER):
        """Combines the img and the img_confidence into a single image.

        Args:
            img (np.ndarray): Label image.
            img_confidence (np.ndarray): Confidence image.
            class_name_mapper (function, optional): Describes how to map the data labels. Defaults to lambdax:str(x).
            colors_mapper (dict, optional): Mapper of labels to colors. Defaults to COLORS_MAPPER.

        Returns:
            np.ndarray: The combined image.
            patches: The legend patches.
        """
        color_img = np.zeros((img.shape[0], img.shape[1], 3))
                
        for i, j in np.ndindex(img.shape):
            color_img[i][j] = colors_mapper[img[i][j]]
        values = np.unique(img)
        
        patches = []
        for value in values:
            color = colors_mapper[value]
            if value==-1:
                label = "Original train data"
            elif value==-2:
                label = "Original test data"
            else:
                label = f"Value region: {class_name_mapper(value)}"

            patches.append(Patch(color=color, label=label))
        
        return color_img, patches
   
    def _generate_encoded_mapping_(self):
        """Generates a mapping of the encoded data to the original data.

        Returns:
            train_mapper (dict): Mapping of the encoded train data to the original train data.
            test_mapper (dict): Mapping of the encoded test data to the original test data.
        """
        train_mapper = {}
        for k in range(len(self.encoded_train)):
            [x, y] = self.encoded_train[k]
            train_mapper[f"{x} {y}"] = k
    
        test_mapper = {}
        for k in range(len(self.encoded_test)):
            [x, y] = self.encoded_test[k]
            test_mapper[f"{x} {y}"] = k
        
        return train_mapper, test_mapper

    def _build_plot_(self):                   
        fig = figure.Figure(figsize = (1, 1))
        ax = fig.add_subplot(111)
        ax.set_axis_off()
        return fig, ax
    
    def _build_canvas_(self, fig, key, controls_key):        
        canvas = self.window[key].TKCanvas
        canvas_controls = self.window[controls_key].TKCanvas
        fig_agg = draw_figure_to_canvas(canvas, fig, canvas_controls)
        return canvas, fig_agg, canvas_controls
    
    def _build_annotation_mapper_(self):
        """ Builds the annotation mapper.
            This is used to display the data point label when hovering over the decision boundary.
        """
        image = OffsetImage(self.X_train[0], zoom=2, cmap="gray")
        label = TextArea("Data point label: None")
        
        annImage = AnnotationBbox(image, (0,0), xybox=(50., 50.), xycoords='data', boxcoords="offset points",  pad=0.1,  arrowprops=dict(arrowstyle="->"))        
        annLabels = AnnotationBbox(label, (0,0), xybox=(50., 50.), xycoords='data', boxcoords="offset points",  pad=0.3,  arrowprops=dict(arrowstyle="->"))

        self.ax.add_artist(annImage)
        self.ax.add_artist(annLabels)
        annImage.set_visible(False)
        annLabels.set_visible(False)

        fig_width, fig_height = self.fig.get_size_inches() * self.fig.dpi
          
        def display_annotation(event):
            """Displays the annotation box when hovering over the decision boundary based on the mouse position."""
            if event.inaxes == None:
                return
        
            j, i = int(event.xdata), int(event.ydata)
            
            # change the annotation box position relative to mouse.
            ws = (event.x > fig_width/2.)*-1 + (event.x <= fig_width/2.) 
            hs = (event.y > fig_height/2.)*-1 + (event.y <= fig_height/2.)
            annImage.xybox = (50. * ws, 50. * hs)
            annLabels.xybox = (-50. * ws, 50. *hs)
            # make annotation box visible
            annImage.set_visible(True)
            # place it at the position of the event scatter point
            annImage.xy = (j, i)
            
            x_data, y_data = find_data_point(i,j)                            
            if x_data is not None:
                annImage.set_visible(True)
                image.set_data(x_data)
            else:
                annImage.set_visible(False)
            
            if y_data is not None:    
                annLabels.set_visible(True)
                annLabels.xy = (j, i)
                label.set_text(f"{y_data}")
            else:
                annLabels.set_visible(False)

            self.fig.canvas.draw_idle()
            
        def find_data_point(i, j):
            # search for the data point in the encoded train data
            if self.img[i][j] == -1:
                k = self.train_mapper[f"{i} {j}"]
                if f"{i} {j}" in self.expert_updates_labels_mapper:
                    l = self.expert_updates_labels_mapper[f"{i} {j}"][0]
                    return self.X_train[k], f"Label {self.Y_train[k]} \nExpert label: {l}"  
                return self.X_train[k], f"Label: {self.Y_train[k]}"
            
            # search for the data point in the encoded test data
            if self.img[i][j] == -2:
                k = self.test_mapper[f"{i} {j}"]
                return self.X_test[k], None      
            
            # search for the data point in the 
            point = None #self.spaceNd[i][j]
            return point, None
                    
        def onclick(event):
            """ Open the data point in a new window when clicked on a pixel in training or testing set based on mouse position."""            
            if event.inaxes == None:
                return
            
            if self.update_labels_by_circle_select:
                onclick_circle_strategy(event)
                return
            
            self.console.log("Clicked on: " + str(event.xdata) + ", " + str(event.ydata))
            j, i = int(event.xdata), int(event.ydata)

            if self.img[i][j] != -1:
                self.console.log("Data point not in training set")
                return
                    
            # check if clicked on a data point that already was updated, then remove the update
            if f"{i} {j}" in self.expert_updates_labels_mapper:
                (_, point) = self.expert_updates_labels_mapper[f"{i} {j}"]
                point.remove()
                del self.expert_updates_labels_mapper[f"{i} {j}"]                    
                self.fig.canvas.draw_idle()
                self.updates_logger.log(f"Removed updates for point: ({j}, {i})")
                return
            
            # if clicked on a data point that was not updated, then add the update
            self.current_selected_point = self.ax.plot(j, i, 'bo', markersize=5)[0]
            
            # disable annotations on hover
            #if self.motion_event_cid is not None:
            #    self.fig.canvas.mpl_disconnect(self.motion_event_cid)
            # disable on click event
            if self.click_event_cid is not None:
                self.fig.canvas.mpl_disconnect(self.click_event_cid)
                self.click_event_cid = None
            
            # enable key press events
            self.key_event_cid = self.fig.canvas.mpl_connect('key_press_event', onkey)
       
            self.fig.canvas.draw_idle()
        
        def onkey(event):                                           
            if self.current_selected_point is None:
                return
            
            if event.key.isdigit():
                self.current_selected_point_assigned_label = int(event.key)
                return
            
            (x, y) = self.current_selected_point.get_data()
            
            if event.key == 'escape' or event.key == 'enter':                            
                self.current_selected_point.remove()       
                
                if event.key == 'escape':                    
                    self.updates_logger.log("Cancelled point move...")                    
                    
                elif event.key == 'enter':
                    # showing the fix of the position                    
                    self.current_selected_point = self.ax.plot(x, y, 'b^')[0]                 
                    if self.current_selected_point_assigned_label is not None:
                        self.updates_logger.log(f"Assigned label: {self.current_selected_point_assigned_label} to point: ({x[0]}, {y[0]})")                        
                        self.expert_updates_labels_mapper[f"{y[0]} {x[0]}"] = (self.current_selected_point_assigned_label, self.current_selected_point)                        
                    
                self.current_selected_point = None
                self.current_selected_point_assigned_label = None
               
                self.motion_event_cid = self.fig.canvas.mpl_connect('motion_notify_event', display_annotation)                
                self.click_event_cid = self.fig.canvas.mpl_connect('button_press_event', onclick)
                self.fig.canvas.mpl_disconnect(self.key_event_cid)  
                self.fig.canvas.draw_idle()  
                return
        
        def onclick_circle_strategy(event):                        
            self.console.log("Clicked on: " + str(event.xdata) + ", " + str(event.ydata))
            x, y = int(event.xdata), int(event.ydata)
            self.current_selected_point = (x,y)
            self.release_event_cid = self.fig.canvas.mpl_connect('button_release_event', onrelease_circle_strategy)
        
        def onrelease_circle_strategy(event):
            if event.inaxes == None:
                return
            self.console.log("Released on: " + str(event.xdata) + ", " + str(event.ydata))
            (x0, y0) = self.current_selected_point
            x1, y1 = int(event.xdata), int(event.ydata)
            (cx, cy) = (x0 + (x1 - x0)/2, y0 + (y1 - y0)/2) # circle center
            r = sqrt(((x1 - x0)/2)**2 + ((y1 - y0)/2)**2)   # circle radius
            self.update_labels_circle = Circle((cx, cy), r, color='black', fill=False)
            self.ax.add_artist(self.update_labels_circle)    
            self.key_event_cid = self.fig.canvas.mpl_connect('key_press_event', onkey_circle_strategy)    
            if self.release_event_cid is not None:
                self.fig.canvas.mpl_disconnect(self.release_event_cid)
            self.fig.canvas.draw_idle()

        def onkey_circle_strategy(event):
            if self.update_labels_circle is None:
                return

            if event.key.isdigit():
                self.current_selected_point_assigned_label = int(event.key)
                return
            
            (x, y) = self.update_labels_circle.get_center()
            r = self.update_labels_circle.get_radius()
            if event.key == "enter" and self.current_selected_point_assigned_label is not None:                                    
                self.updates_logger.log(f"Assigned label: {self.current_selected_point_assigned_label} to circle: center ({x}, {y}), radius ({r})")
                positions = find_points_in_circle((x,y), r)
                for pos in positions:
                    point = self.ax.plot(pos[0], pos[1], 'b^')[0]   
                    self.expert_updates_labels_mapper[f"{pos[1]} {pos[0]}"] = (self.current_selected_point_assigned_label, point)                        
                           
            if event.key == "backspace":
                self.updates_logger.log(f"Removing changes in circle: center ({x},{y}), radius ({r})")
                positions = find_points_in_circle((x,y), r)
                for pos in positions:
                    if f"{pos[1]} {pos[0]}" in self.expert_updates_labels_mapper:
                        self.expert_updates_labels_mapper[f"{pos[1]} {pos[0]}"][1].remove()
                        del self.expert_updates_labels_mapper[f"{pos[1]} {pos[0]}"]
                
            self.update_labels_circle.remove()
            self.update_labels_circle = None
            self.fig.canvas.mpl_disconnect(self.key_event_cid)  
            self.fig.canvas.draw_idle()
        
        def find_points_in_circle(circle_center, circle_radius):
            (cx, cy) = circle_center
            initial_x, initial_y = int(cx - circle_radius), int(cy - circle_radius)
            final_x, final_y = int(cx + circle_radius) + 1, int(cy + circle_radius) + 1
            initial_x = 0 if initial_x < 0 else initial_x
            initial_y = 0 if initial_y < 0 else initial_y
            final_x = self.img.shape[0] if final_x > self.img.shape[0] else final_x
            final_y = self.img.shape[0] if final_y > self.img.shape[0] else final_y
            
            positions = []
            for x in range(initial_x, final_x):
                for y in range(initial_y, final_y):
                    if (self.img[y, x] == -1) and ((x - cx)**2 + (y - cy)**2 <= circle_radius**2):
                        positions.append((x,y))
            return positions
            
                    
        self.motion_event_cid = self.fig.canvas.mpl_connect('motion_notify_event', display_annotation)           
        self.click_event_cid = self.fig.canvas.mpl_connect('button_press_event', onclick)
       
    def plot_data_point(self, data, label):
        """Plots the data point in a new window.

        Args:
            data (np.ndarray): the data point to be displayed
            label (str): the label of the data point
        """
        img = Image.fromarray(np.uint8(data * 255))
        img.thumbnail((100, 100), Image.ANTIALIAS)
        img.show(title=f"Data point label: {label}")
    
    def draw_dbm_img(self):             
        # update the figure
        self.ax.set_title("Decision Boundary Mapper")    
        img = np.zeros((self.img.shape[0], self.img.shape[1], 4))
        img[:,:,:3] = self.color_img
        img[:,:,3] = self.img_confidence    
        self.axes_image = self.ax.imshow(img)                                    
                    
        # draw the figure to the canvas
        self.fig_agg = draw_figure_to_canvas(self.canvas, self.fig, self.canvas_controls)    
        self.window.refresh()
        
    def compute_classifier_metrics(self):
        self.console.log("Evaluating classifier...")
        loss, accuracy = self.dbm_model.classifier.evaluate(self.X_test, self.Y_test, verbose=0)        
        self.console.log(f"Classifier Accuracy: {(100 * accuracy):.2f}%  Loss: {loss:.2f}")
        
        path = os.path.join(self.save_folder, CLASSIFIER_PERFORMANCE_HISTORY_FILE)
        
        with open(path, "a") as f:
            time = datetime.now().strftime("%D %H:%M:%S")
            message = f"Accuracy: {(100 * accuracy):.2f}%  Loss: {loss:.2f}"
            f.write(f"{time} {message}\n")
            
        self.window["-CLASSIFIER ACCURACY-"].update(f"Classifier Accuracy: {(100 * accuracy):.2f} %  Loss: {loss:.2f}")
    
    def handle_compute_inverse_projection_errors_event(self, event, values):
        self.window['-COMPUTE INVERSE PROJECTION ERRORS-'].update(visible=False, disabled=True)        
        self.updates_logger.log("Computing inverse projection errors, please wait...")
        possible_path = os.path.join(self.save_folder, "inverse_projection_errors.npy")
        # try to get projection errors from cache first
        if os.path.exists(possible_path):
            self.inverse_projection_errors = np.load(possible_path)
        else:
            self.inverse_projection_errors = self.dbm_model.generate_inverse_projection_errors(save_folder=self.save_folder, resolution = len(self.img))

        self.updates_logger.log("Inverse projection errors computed!")
        self.window['-SHOW INVERSE PROJECTION ERRORS-'].update(visible=True)
            
    def _set_loading_proj_errs_state_(self):
        self.window['-COMPUTE PROJECTION ERRORS INTERPOLATION-'].update(visible=False, disabled=True)        
        self.window['-COMPUTE PROJECTION ERRORS INVERSE PROJECTION-'].update(visible=False, disabled=True)        
        self.updates_logger.log("Computing projection errors, please wait...")
        
    def handle_compute_projection_errors_event(self, event, values):
        self._set_loading_proj_errs_state_()
        if event == "-COMPUTE PROJECTION ERRORS INTERPOLATION-":
            possible_path = os.path.join(self.save_folder, "projection_errors_interpolated.npy")
            # try to get projection errors from cache first
            if os.path.exists(possible_path):
                self.projection_errors = np.load(possible_path)
            else:
                self.projection_errors = self.dbm_model.generate_projection_errors(use_interpolation=True, save_folder=self.save_folder)
        elif event == "-COMPUTE PROJECTION ERRORS INVERSE PROJECTION-":            
            possible_path = os.path.join(self.save_folder, "projection_errors_inv_proj.npy")
            # try to get projection errors from cache first
            if os.path.exists(possible_path):
                self.projection_errors = np.load(possible_path)
            else:
                self.projection_errors = self.dbm_model.generate_projection_errors(use_interpolation=False, save_folder=self.save_folder)        
        
        self.updates_logger.log("Finished computing projection errors.")
        self.window['-SHOW PROJECTION ERRORS-'].update(visible=True)           

    def handle_checkbox_change_event(self, event, values):
        show_color_map = values["-SHOW DBM COLOR MAP-"]
        show_confidence = values["-SHOW DBM CONFIDENCE-"]
        show_inverse_projection_errors = values["-SHOW INVERSE PROJECTION ERRORS-"]
        show_projection_errors = values["-SHOW PROJECTION ERRORS-"]
        show_labels_changes = values["-SHOW LABELS CHANGES-"]
        
        color_img = np.zeros((self.img.shape[0], self.img.shape[1], 3))
        alphas = 1 + np.zeros((self.img.shape[0], self.img.shape[1]))
        img = np.zeros((self.img.shape[0], self.img.shape[1], 4))
        
        if show_color_map:
            color_img = self.color_img
        
        if show_confidence:
            alphas = alphas * self.img_confidence
        if show_inverse_projection_errors:
            alphas = alphas * (1 - self.inverse_projection_errors)
            
        if show_projection_errors and self.projection_errors is not None:
            alphas = alphas * (1 - self.projection_errors)
        
        img[:, :, :3] = color_img
        img[:, :, 3] = alphas                     
        
        if hasattr(self, "axes_image"):
            self.axes_image.remove()
        self.axes_image = self.ax.imshow(img) 
        
        if hasattr(self, "ax_labels_changes") and self.ax_labels_changes is not None:            
            self.ax_labels_changes.set_visible(False)
            #for point in self.ax_labels_changes:                
            #    point.remove()
            self.ax_labels_changes = None
        
        positions_x, positions_y, alphas = self.positions_of_labels_changes   
        if show_labels_changes and len(positions_x) > 0 and len(positions_y) > 0  and len(alphas) > 0:                             
            self.ax_labels_changes = self.ax.scatter(positions_x, positions_y, c='green', marker='^', alpha=alphas)   
        
        self.fig.canvas.draw_idle()
    
    def handle_circle_selecting_labels_change_event(self, event, values):
        self.update_labels_by_circle_select = values["-CIRCLE SELECTING LABELS-"]
        
    def handle_apply_changes_event(self, event, values):
        num_changes = len(self.expert_updates_labels_mapper)
        if num_changes == 0:
            self.updates_logger.log("No changes to apply")
            return
        
        if num_changes < 10:
            self.updates_logger.log("Less than 10 changes to apply, please apply more changes")
            return
        
        """
        if values["-EPOCHS-"].isdigit():
            epochs = int(values["-EPOCHS-"])
        else:
            epochs = 2
            self.window["-EPOCHS-"].update(epochs)
        """
        epochs = 2
        
        self.console.log("Transforming changes...")
        Y, label_changes = self.transform_changes()
        
        self.console.log("Saving changes to a local folder...")
        self.save_changes(self.save_folder, label_changes=label_changes)
        
        self.updates_logger.log("Applying changes... This might take a couple of seconds, after this the window will be closed")
                
        self.dbm_model.refit_classifier(self.X_train, Y, save_folder=os.path.join(self.save_folder, "refit_classifier"), epochs = epochs)
        
        # Updating the main GUI with the new model                
        dbm_info = self.dbm_model.generate_boundary_map(
            self.X_train, 
            Y, 
            self.X_test, 
            self.Y_test, 
            resolution=len(self.img),
            use_fast_decoding=True,
            load_folder=self.save_folder,
            projection=self.projection_technique                                        
        )        
            
        img, img_confidence, encoded_training_data, encoded_testing_data, training_history = dbm_info
        self.initialize(dbm_model = self.dbm_model,
                            img = img,
                            img_confidence = img_confidence,
                            encoded_train = encoded_training_data, 
                            encoded_test = encoded_testing_data,
                            X_train = self.X_train, 
                            Y_train = Y,
                            X_test = self.X_test,
                            Y_test = self.Y_test,                            
                            save_folder=self.save_folder,
                            projection_technique=self.projection_technique,
                        )
        
        self.draw_dbm_img()   
        self.handle_checkbox_change_event(event, values)      
        self.compute_classifier_metrics()
        
        self.updates_logger.log("Changes applied successfully!")     
            
    def save_changes(self, folder:str="tmp", label_changes={}):
        if not os.path.exists(folder):
            os.path.makedirs(folder)                        
        
        with open(os.path.join(folder, LABELS_CHANGES_FILE), "a") as f:
            f.write("\n" + "-" * 50 + "\n")
            json.dump(label_changes, f, indent=2)
    
    def transform_changes(self):
        """
        Returns:
            position_changes (dict): a dictionary with old position as key and the new position as value
            label_changes (dict): a dictionary with old and new positions as key and the new label as value
        """     
        Y = np.copy(self.Y_train) 
        labels_changes = {}
        
        pos_x, pos_y, alphas = self.positions_of_labels_changes
        
        ALPHA_DECAY_ON_UPDATE = 0.05        
        alphas = [alpha - ALPHA_DECAY_ON_UPDATE if alpha > ALPHA_DECAY_ON_UPDATE else ALPHA_DECAY_ON_UPDATE for alpha in alphas]
        
        for pos in self.expert_updates_labels_mapper:
            k = self.train_mapper[pos]
            pos_x.append(int(pos.split(" ")[1]))
            pos_y.append(int(pos.split(" ")[0]))
            alphas.append(1)
            y = self.expert_updates_labels_mapper[pos][0]
            Y[k] = y
            labels_changes[pos] = self.expert_updates_labels_mapper[pos][0]
        
        self.positions_of_labels_changes = (pos_x, pos_y, alphas)
            
        return Y, labels_changes
    
    def handle_show_classifier_performance_history_event(self, event, values):
        path = os.path.join(self.save_folder, CLASSIFIER_PERFORMANCE_HISTORY_FILE)
        if not os.path.isfile(path):
            return
        
        times, accuracies, losses = [], [], []
        with open(path, "r") as f:
            for line in f.readlines():                
                line = line.strip()
                if len(line) == 0:
                    continue
                _, time, _, acc, _, _, loss = line.replace("\n", "").replace("%", "").split(" ")                        
                times.append(time)
                accuracies.append(float(acc))
                losses.append(float(loss))
        
        fig, (ax1, ax2) = plt.subplots(1,2, figsize=(20,10))
        
        for tick in ax1.get_xticklabels():
            tick.set_rotation(45)
        for tick in ax2.get_xticklabels():
            tick.set_rotation(45)
        
        ax1.set_title("Classifier accuracy history")
        ax2.set_title("Classifier loss history")
        ax1.set_xlabel("Time")
        ax2.set_xlabel("Time")
        ax1.set_ylabel("Accuracy (%)")
        ax2.set_ylabel("Loss")
        
        ax1.plot(times, accuracies)
        ax2.plot(times, losses)
        plt.show()
        